package {
	import flare.animate.Transitioner;
	import flare.display.TextSprite;
	import flare.scale.ScaleType;
	import flare.scale.TimeScale;
	import flare.util.Dates;
	import flare.util.palette.ColorPalette;
	import flare.util.Shapes;
	import flare.util.Strings;
	import flare.vis.Visualization;
	import flare.vis.controls.ClickControl;
	import flare.vis.controls.HoverControl;
	import flare.vis.controls.TooltipControl;
	import flare.vis.data.Data;
	import flare.vis.data.DataSprite;
	import flare.vis.data.NodeSprite;
	import flare.vis.events.SelectionEvent;
	import flare.vis.events.TooltipEvent;
	import flare.vis.operator.encoder.ColorEncoder;
	import flare.vis.operator.filter.VisibilityFilter;
	import flare.vis.operator.label.StackedAreaLabeler;
	import flare.widgets.ProgressBar;
	import flare.widgets.SearchBox;
	
	import flash.display.Sprite;
	import flash.events.Event;
	import flash.events.KeyboardEvent;
	import flash.events.MouseEvent;
	import flash.filters.DropShadowFilter;
	import flash.geom.Rectangle;
	import flash.text.TextField;
	import flash.text.TextFieldType;
	import flash.text.TextFormat;
	import flash.ui.Keyboard;
	
	import lastfmvis.LastfmXMLLoader;
	import lastfmvis.StackedAreaLayout;
	
	import mx.core.Application;
	import mx.core.UIComponent;

	[Event(name="VIS_LOADED")]

	public class Last_fm_vis extends UIComponent
	{
		public static const VIS_LOADED:String = "vis loaded";
		
		// Data
		// EXAMPLE DATA!  THIS GETS OVERWRITEN BY REAL DATA	
		private var rawData:Array = [
			{ date:1221393600, artist:"Squarepusher", playcount:33 },
			{ date:1221393600, artist:"Air", playcount:16 },
			{ date:1221393600, artist:"Portishead", playcount:14 },
			{ date:1221393600, artist:"Ratatat", playcount:14 },
			{ date:1220788800, artist:"Feist", playcount:22 },
			{ date:1220788800, artist:"Tricky", playcount:12},
			{ date:1220788800, artist:"Ratatat", playcount:11},
			{ date:1220788800, artist:"Modest Mouse", playcount:11}
		];

		// EXAMPLE DATA
		// same as rawData, but for tags
		private var rawTagData:Array = [
			{ date:1221393600, tag:"Rock", playcount:45 },
			// ...
		];
		
		private var TAG_TYPE:Number = 1;
		private var ARTIST_TYPE:Number = 0;
		
		private var allRawData:Array = [
		];
		
		// Mapping from artist to tag
		private var artistTags:Object = {};
		
		// Same data as rawTagData array, but in hash form.
		// tagCounts = { Rock: { 1221393600: 45}}
		// (used as intermediary format while assembling tag counts)
		private var tagCounts:Object = {};
		
		// The tag data once it's been rotated for visualization
//		private var rotatedTagData:Data;
		
		// The artist data once it's been rotated for visualization 
//		private var rotatedArtistData:Data;

		private var rawChartData:Array = [];
		private var weeks:Array = [];
		private var chartData:Data;
		private var removedData:Object = {};
		
		// Alex's API key
		private var API_KEY:String = "c669ec5933241d05076deab833cf48b8";
		// Nick's API key
		//private var API_KEY:String = "38d9ae932c69a89d269c9d4119e7363c";
		
		// Username vars
		private var usernameEntryFormat:TextFormat = new TextFormat("Trebuchet MS",25,0,true,
			null, null, null, null, "left");
		private var usernameEntry:TextField;
		private var username:String = "";	
		private var usernameLabel:TextSprite = new TextSprite("Username", usernameEntryFormat);
	
		// Progress bar
		private var progressBar:ProgressBar = new ProgressBar(250,25);

		// Property to base stacks of data on, artist, tag, etc
		private var filterProperty:String;
		
		// Search box
		private var search:SearchBox;
		//private var fmt:TextFormat = new TextFormat("Helvetica,Arial",16,0,true);
		private var fmt:TextFormat = new TextFormat("Arial",16,0,true);
		
		// Visualiation
		private var vis:Visualization;
		private var _bounds:Rectangle;
		private var _t:Transitioner;
		private var _dur:Number = 1.25;	// Animation duration
		private var _timeRangeDur:Number = 0.5;
		
		private var _dates:Array = [];
		
		// Keep track of which weeks we have, so we can figure out which ones we don't
		private var foundWeeks:Object = {};
		
		// The tag that is currently being viewed; for artist view.
		private var currentTag:String;
		

		// Query variables
		private var query:Array;
		private var exact:Boolean; // Exact search requested

		// Size of component
		private var totalWidth:Number = 800;
		private var totalHeight:Number = 600;
		
		// Box in right-hand corner which shows  tag info
		private var tagInfoBox:Sprite;

		public function get dates():Array { return _dates; }

		/**
		 * Runs the program.
		 */
		public function Last_fm_vis()
		{
			init();
		}
		
		private function init():void {
			width = totalWidth;
			height = totalHeight;
			
			initUsername();
			initProgressBar();
		}
		
		private function initUsername():void {
			// initialize the usernameEntry text field
			usernameEntry = new TextField();
			usernameEntry.type = TextFieldType.INPUT;
			usernameEntry.defaultTextFormat = usernameEntryFormat;
			usernameEntry.width = 250;
			usernameEntry.height = usernameEntryFormat.size.valueOf() + 7;
			usernameEntry.x = (this.width - usernameEntry.width) / 2;
			usernameEntry.y = (this.height - usernameEntry.height) / 2;
			usernameEntry.border = true;
			usernameEntry.borderColor = 0x000000;
			usernameEntry.visible = true;
			usernameEntry.text = "";
			
			// Initialize the label for the field
			usernameLabel.x = usernameEntry.x;
			usernameLabel.y = usernameEntry.y - usernameEntry.height;
			usernameLabel.visible = true;
			
			addChild(usernameEntry);
			addChild(usernameLabel);
			
			// For some reason this command isn't working; it should automatically
			// select the text field. Something to figure out later...
			//stage.focus = usernameEntry;
			
			usernameEntry.addEventListener(KeyboardEvent.KEY_DOWN, usernameKeyDown);		
		}
		
		private function initProgressBar():void {
			progressBar.x = (this.width - progressBar.width) / 2;
			progressBar.y = (this.height - progressBar.height) / 2;
			progressBar.visible = false;
			progressBar.setMessage("Contacting Last.fm...")
			progressBar.bar.filters = [new DropShadowFilter(1)];
			
			addChild(progressBar);
		}
		
		private function usernameKeyDown(evt:KeyboardEvent):void {
			if(evt.keyCode == Keyboard.ENTER)
			{
				usernameEntry.removeEventListener(KeyboardEvent.KEY_DOWN, usernameKeyDown);
				usernameEntry.visible = false;
				usernameLabel.visible = false;
				username = usernameEntry.text;
				usernameRetrieved();
			}
			return;
		}
		
	//	private function usernameRetrieved(evt:KeyboardEvent):void {
		private function usernameRetrieved():void {
//			trace("In key up");
//			this.removeEventListener(KeyboardEvent.KEY_UP, usernameRetrieved);
			progressBar.visible = true;
			getWeeks();
		}
		
		/**
		 * Gets the weeks of available data for a given user, so we know
		 * which week have listening data in the first place.
		 */
		private function getWeeks():void {
			var xmlLoader:LastfmXMLLoader = new LastfmXMLLoader(API_KEY);
			xmlLoader.onComplete = function(xmlData:Array):void {
				var weeksXML:XML = xmlData[0];
				for each (var week:XML in weeksXML..chart) {
					var start:Number = Number(week.@from.toString());
					// Arbitrary cut-off to avoid endless wait times while debugging:
//					if (start > 1217764800)
					// Download last year
					if (start > 1192363200)
						weeks.push([start, Number(week.@to.toString())]);
				}
				for each (var range:Array in weeks) {
					var newDate:Date = new Date(Number(range[0]) * 1000);
					_dates.push(newDate);
					foundWeeks[newDate] = true;
				}
				getListeningData();
			};
			xmlLoader.retrieveAvailableWeeks(username);
		}
				
		/**
		 * Gets the weekly artist listening data for the weeks
		 * gotten by getWeeks()
		 */
		private function getListeningData():void {
			rawData = [];
			progressBar.setMessage("Loading week 0 of " + weeks.length);

			var xmlLoader:LastfmXMLLoader = new LastfmXMLLoader(API_KEY);
			xmlLoader.retrieveWeeklyArtistListenData(username, weeks);
			xmlLoader.onProgress = function(progress:Number, total:Number):void {
				progressBar.progress = progress/total;
				progressBar.setMessage("Loading week " + progress + " of " + total);
			}
			xmlLoader.onComplete = function(xmlData:Array):void {
				for each (var listenData:XML in xmlData) {
					var fromAttr:XMLList = listenData..weeklyartistchart[0].@from;
					var fromDate:Number = Number(fromAttr.toString());
					for each (var artist:XML in listenData..artist) {
						var artistName:String = trim(artist.name.toString());
						// Convert timestamp in XML to AS3 Date object
						rawData.push( { date:new Date(fromDate * 1000), 
							  			artist:artistName, 
							  			playcount:Number(artist.playcount.toString())
						}); 
					}
				}
				weeklyDataLoaded();
			}
		}
		
		/**
		 *	Called when all the weekly listening data has been downloaded.
		 */
		public function weeklyDataLoaded():void {
			//progressBar.visible = false;
			progressBar.progress = 0;
			progressBar.setMessage("Retrieving artist tags...");
			// Tally up unique artists and get their tags
			var artists:Array = [];
			var foundArtists:Object = {};
			for each (var row:Object in rawData) {
				if (! foundArtists[row.artist]) {
					artists.push(row.artist);
					foundArtists[row.artist] = true;
				}
			}
			var xmlLoader:LastfmXMLLoader = new LastfmXMLLoader(API_KEY);
			xmlLoader.onProgress = function(progress:Number, total:Number):void {
				progressBar.progress = progress/total
				progressBar.setMessage("Retrieving artist " + progress + " of " + total);	
			};
			xmlLoader.onComplete = function(xmlData:Array):void {
				for each (var artistTagData:XML in xmlData) {
					if(artistTagData != null)
					{
						var artist:String = artistTagData.toptags[0].@artist.toString();
						if (artistTagData..tag[0] != null) {
							var tag:XML = artistTagData..tag[0].name[0];
							artistTags[artist] = tag.toString(); 
						}
					}
				}
				progressBar.visible = false;
				countTagListens();
			}
			xmlLoader.retrieveTagsForArtists(artists);
			
			printArray(rawData);
/*			rawChartData =  reshape(rawData, ["artist"], "date", "playcount", _dates, false);
			//printArray(data_rotated);
			printArray(rawChartData);
			chartData = Data.fromArray(rawChartData);
			
			visualize(chartData);
			addControls();
			layout();

			fireEvent(VIS_LOADED);
*/
		}
		
		
		
		/**
		 * Tallies up the listens for each tag for each week once all the tags
		 * and artists have been retrieved.
		 */
		private function countTagListens():void {
			for each (var row:Object in rawData) {
				var artist:String = row["artist"];
				var date:Date = row["date"];
				if (tagCounts[artistTags[artist]]) {
					if (tagCounts[artistTags[artist]][date]) {
						tagCounts[artistTags[artist]][date] += row["playcount"];
					} else {
						tagCounts[artistTags[artist]][date] = row["playcount"];
					}
				} else {
					tagCounts[artistTags[artist]] = {};
					tagCounts[artistTags[artist]][date] = row["playcount"];
				}
			}
			rawTagData = [];
			for (var tagStr:String in tagCounts) {			
				for (var dateStr:String in tagCounts[tagStr]) {
					rawTagData.push({ date:new Date(dateStr), tag:tagStr, playcount:tagCounts[tagStr][dateStr] });
					trace("added " + tagStr + " for date: " + new Date(dateStr) + " count: " + tagCounts[tagStr][dateStr]);
				}
			}
			
			allRawData = [];
			for each (row in rawData) {
				allRawData.push({ date:row.date, name:row.artist, playcount:row.playcount, type:ARTIST_TYPE });
			} 
			for each (row in rawTagData) {
				allRawData.push({ date:row.date, name:row.tag, playcount:row.playcount, type:TAG_TYPE });
			}
			
			/* Tag Visualization: */
//			var tag_data_rotated:Array = reshape(rawTagData, ["tag"], "date", "playcount", _dates, false);
			/*		Artist visualization: */			
//			var artist_data_rotated:Array = reshape(rawData, ["artist"], "date", "playcount", _dates, false);
			
			// Finding missing weeks and add them, so we know there's no data for that week
			var newWeeks:Array = [];
			for each (var aWeek:Date in _dates) {
				var nextWeek:Date = Dates.addDays(aWeek, 7);
				if (!foundWeeks[nextWeek]) {
					newWeeks.push(nextWeek);
				}
			}
			_dates = _dates.concat(newWeeks);
			_dates.sortOn("time");
			var combined_data_rotated:Array = reshape(allRawData, ["name", "type"], "date", "playcount", _dates, false);

//			rotatedTagData = Data.fromArray(tag_data_rotated);
//			rotatedArtistData = Data.fromArray(artist_data_rotated);
			var combinedData:Data = Data.fromArray(combined_data_rotated);
			initVisualization(combinedData);
			// Start by visualizing tag data: Not using visualizeTags() becuse
			// that update will conflict with the layout()-called update.			
			filterProperty="tag";
			addControls();
			layout();
			
			fireEvent(VIS_LOADED);
		}
		
		private function initVisualization(data:Data):void {
			// Define the visualization		
			vis = new Visualization(data);
			
			data.nodes.sortBy("+data.playcount");
			data.nodes.setProperties({
				shape: Shapes.POLYGON,
				lineColor: 0,
				fillValue: 1,
				fillSaturation: 0.5,
				fillAlpha: 0.6
			});
			
			// first, set the visibility according to the query
			vis.operators.add(new VisibilityFilter(filter));
			vis.operators[0].immediate = true; // filter immediately!
			
			vis.operators.add(new StackedAreaLayout(_dates, 0));
			// Minimum width of stack is 2 pixels
			//vis.operators[1].threshold = 2;
			vis.operators[1].threshold = 1;
			vis.operators[1].colScale.labelFormat = "MM/dd/yy";
			
			// Labeler
			vis.operators.add(new StackedAreaLabeler("data.name"));
			// Column to label
			vis.operators[2].columnIndex = 1;
			
			vis.operators.add(new ColorEncoder("data.name", Data.NODES, "fillColor",
							  ScaleType.CATEGORIES, ColorPalette.category(data.length/2,null,0.6)));
			
			vis.xyAxes.yAxis.verticalAnchor = TextSprite.TOP;
			vis.xyAxes.yAxis.horizontalAnchor = TextSprite.RIGHT;
			vis.xyAxes.yAxis.labelOffsetX = 50;  // offset labels to the right
			vis.xyAxes.yAxis.lineCapX1 = 15; // extra line length to the left
			vis.xyAxes.yAxis.lineCapX2 = 50; // extra line length to the right
			vis.xyAxes.showBorder = false;
			addChild(vis);
			
			// add mouse-over highlight
			vis.controls.add(new HoverControl(NodeSprite,
			    // move highlighted node to be drawn on top
				HoverControl.MOVE_AND_RETURN,
				// highlight node to full saturation
				function(e:SelectionEvent):void {
					e.node.props.alpha = e.node.fillAlpha;
					e.node.props.saturation = e.node.fillSaturation;
					e.node.fillAlpha = 1;
					e.node.fillSaturation = 1;
				},
				// return node to previous saturation
				function(e:SelectionEvent):void {
					e.node.fillAlpha = e.node.props.alpha;
					e.node.fillSaturation = e.node.props.saturation;
				}
			));
				
			// add filter on click
			vis.controls.add(new ClickControl(NodeSprite, 1,
				// set search query to the artist name
				function(e:SelectionEvent):void {
					exact = true; // force an exact search
					search.query = e.node.data["name"];
				}
			));
			
			// add artist view on double-click
			vis.controls.add(new ClickControl(NodeSprite, 2,
				// Drill down to artists-in-tag view
				function(e:SelectionEvent):void {
					// TODO: some text display of the tag and a back button
					if(filterProperty == "tag")
					{
						currentTag = e.node.data["name"];
						visualizeArtists();
					}
				}
			));
			
			// add tooltips
			vis.controls.add(new TooltipControl(NodeSprite, null,
			    // update on both roll-over and mouse-move
				updateTooltip, updateTooltip));	
		}

		/**
		 * Creates the tooltip text
		 */
		private function updateTooltip(e:TooltipEvent):void
		{
			// get current year value from axes, and map to data
			var date:Date =
				vis.xyAxes.xAxis.value(vis.mouseX, vis.mouseY) as Date;
			
			// Find the closest available date to the cursor
			date = Dates.addDays(date, 4);
			date = Dates.roundTime(date,Dates.WEEKS);
			// All times are at 12:00 PM UTC
			date.setUTCHours(12);

			var def:Boolean = (e.node.data[date] != undefined);
			
			if (!def) {
				// Sometimes dates are stored at 4:00 .... weird GMT/Daylight Savings Issue
				date = Dates.addHours(date, -1);
			}
			
			// Tooltip for artist view
			if(filterProperty == "artist")
			{
				TextSprite(e.tooltip).htmlText = Strings.format(
					"<b>Artist</b>: {0}<br/><b>Playcount</b>: {1:g}<br/><b>Week beginning:</b> {2:MM/dd/yy}",
					e.node.data.name, e.node.data[date], date);
			}
			// Tooltip for tag view
			else if(filterProperty == "tag")
			{
				TextSprite(e.tooltip).htmlText = Strings.format(
					"<b>Tag</b>: {0}<br/><b>Playcount</b>: {1:g}<br/><b>Week beginning:</b> {2:MM/dd/yy}",
					e.node.data.name, e.node.data[date], date);
			}
		}
		
		/**
		 * Create the stacked graph visualization for tag-based data
		 */
		private function visualizeTags():void {
			filterProperty = "tag";
			// Sort by size, descending.
/*			rotatedTagData.nodes.sortBy("-data.max");
			rotatedTagData.nodes.setProperties({
				shape: Shapes.POLYGON,
				lineColor: 0,
				fillValue: 1,
				fillSaturation: 0.5
			});
			vis.data = rotatedTagData; 
//			vis.labels = null;
			vis.operators[2].source = "data.tag";
			vis.operators[2].setup();
			vis.operators[3].source = "data.tag";
			vis.operators[3].setup();
//			vis.xyAxes.yAxis.labelOffsetX = 50; */
			_t = vis.update(_timeRangeDur);
			_t.play();
			search.query = "";
		}
		
		/**
		 * Create the stacked graph visualization of artist data
		 */
		private function visualizeArtists():void {
			// Clear search box, in case you were viewing a specific tag
			tagInfoBox = new Sprite();
			var legendItem:TextSprite = new TextSprite("back to all tags", fmt);
			tagInfoBox.addChild(new TextSprite("Viewing tag: "+ currentTag, usernameEntryFormat));
			tagInfoBox.addChild(legendItem);
			TextSprite(tagInfoBox.getChildAt(0)).y = 10;
			TextSprite(tagInfoBox.getChildAt(0)).bold = false;
			TextSprite(tagInfoBox.getChildAt(0)).render();
			legendItem.buttonMode = true;
			legendItem.bold = true;
			legendItem.y = 40;
			legendItem.render();
			legendItem.addEventListener(MouseEvent.CLICK,
					function(e:MouseEvent):void {
//						tagInfoBox.removeChildAt(0);
//						tagInfoBox.removeChildAt(1);
						removeChild(tagInfoBox);
						search.query = "";
						visualizeTags();
					}
				);
			tagInfoBox.x = 400;
			addChild(tagInfoBox);
			filterProperty = "artist";
			// Sort by size, descending.
/*			rotatedArtistData.nodes.sortBy("-data.max");
			rotatedArtistData.nodes.setProperties({
				shape: Shapes.POLYGON,
				lineColor: 0,
				fillValue: 1,
				fillSaturation: 0.5
			});
			vis.data = rotatedArtistData;
			// Seem to need to clear off old labels ... go figure.
			vis.labels = null;
			vis.operators[2].source = "data.artist";
			vis.operators[2].setup();
			// Column to label
			vis.operators[3].source = "data.artist";
			vis.operators[3].setup();
//			vis.xyAxes.yAxis.labelOffsetX = 50; */
			_t = vis.update(_timeRangeDur);
			_t.play();
			search.query = "";
		}

		public function changeMinDate(ind:int):void {
			var l:StackedAreaLayout = vis.operators[1];
			var dateScale:TimeScale = vis.operators[1].colScale;		
			if(dateScale.min != _dates[ind])
			{
				var i:uint = 0;
				var j:uint = 0;
				for(j=0; j<vis.data.nodes.length; j++)
				{
					var ds:DataSprite = vis.data.nodes[j];
					var hash:String = "";
					if(dateScale.min <= _dates[ind])
					{
						for (i=0; i<ind; i++)
						{
							if(ds.data[_dates[i]] != null)
							{
								hash = j.toString() + _dates[i].toString();
								removedData[hash] = ds.data[_dates[i]];
								ds.data[_dates[i]] = null;
							}
						}
					}
					else
					{
						i=ind;					
						while(_dates[i] >= _dates[ind] && _dates[i] <= dateScale.min)
						{
							if(ds.data[_dates[i]] == null)
							{
								hash = j.toString() + _dates[i].toString();
								ds.data[_dates[i]] = removedData[hash];
								delete removedData[hash];
							}
							i++;
						}
					}
				}
				
				l.actualColMin = ind;	
				//l.columns = _dates.slice(ind,_dates.length);
				dateScale.min = _dates[ind];
				_t = vis.update(_timeRangeDur);
				_t.play();
			}
		}

		public function changeMaxDate(ind:int):void {
			var l:StackedAreaLayout = vis.operators[1];
			var dateScale:TimeScale = vis.operators[1].colScale;		
			if(dateScale.max != _dates[ind])
			{
				var i:uint = 0;
				var j:uint = 0;
				for(j=0; j<vis.data.nodes.length; j++)
				{
					var ds:DataSprite = vis.data.nodes[j];
					var hash:String = "";
					if(dateScale.max >= _dates[ind])
					{
						for (i=_dates.length-1; i>ind; i--)
						{
							if(ds.data[_dates[i]] != null)
							{
								hash = j.toString() + _dates[i].toString();
								removedData[hash] = ds.data[_dates[i]];
								ds.data[_dates[i]] = null;
							}
						}
					}
					else
					{
						i=ind;					
						while(_dates[i] <= _dates[ind] && _dates[i] >= dateScale.max)
						{
							if(ds.data[_dates[i]] == null)
							{
								hash = j.toString() + _dates[i].toString();
								ds.data[_dates[i]] = removedData[hash];
								delete removedData[hash];
							}
							i--;
						}
					}
				
					//l.columns = _dates.slice(ind,_dates.length);
				}
				
				l.actualColMax = ind;				
				
				dateScale.max = _dates[ind];
				_t = vis.update(_timeRangeDur);
				_t.play();
			}
		}
		
		public function testRemove():void {
			trace("In testRemove...");
//			var chartNodes:DataList = chartData.nodes;
//			var tempArray:Array = chartData.nodes.toDataArray();
//			removedData.add(chartNodes.list[0]);
//			vis.data.remove(temp[0]);
			var dateScale:TimeScale = vis.operators[1].colScale;
			trace(vis.operators[1].scale.dataMin + " " + vis.operators[1].scale.dataMax);
			trace(dateScale.min + " " + dateScale.max);
			trace(dateScale.scaleMin + " " + dateScale.scaleMax);
			dateScale.min = _dates[3];
			_t = vis.update(_timeRangeDur);
			_t.play();
		}
		public function testAdd():void {
			trace("In testAdd...");
//			var temp:DataSprite = removedData.toDataArray()[0];
//			removedData.remove(temp);
//			vis.data.addNode(temp);
			var dateScale:TimeScale = vis.operators[1].colScale;
			dateScale.min = _dates[0];
			_t = vis.update(_timeRangeDur);
			_t.play();
			
			var scaleValues:Array = vis.operators[1].colScale.values();
			for each (var d:Date in scaleValues) trace(d);
		}
		public function testData():void {
			var temp:Array = chartData.nodes.toDataArray();
			trace("rawChartData[0][_dates[0]]: " + rawChartData[0]["artist"]);
			trace("nodes.length: " + temp.length + "\ntemp[0]: " + temp[15]["artist"]);
		}

		/** Filter function for determining visibility. */
		private function filter(d:DataSprite):Boolean
		{
			var prop:String = String(d.data["name"]);
			// Filter based on double-clicked tag
			if (filterProperty == "artist" && d.data["type"] != ARTIST_TYPE) {
				return false;
			}
			if (filterProperty == "tag" && d.data["type"] != TAG_TYPE) {
				return false;
			}
			if (filterProperty == "artist" && currentTag != null) {
				if (artistTags[prop] != currentTag)
					return false;
			}
			if (!query || query.length==0) {
				return true;
			} else {
				var s:String = prop.toLowerCase();
				for each (var q:String in query) {
					var len:int = q.length;
					if (len == 0) continue;
					// Search anywhere in string, not just beginning
					if (!exact && (s.search(q) != -1)) return true;
					if (exact && q==s) return true;
				}
				return false;
			}
		}

		/** Callback for filter events. */
		private function onFilter(evt:Event=null):void
		{
			query = search.query.toLowerCase().split(/\|/);
			if (query.length==1 && query[0].length==0) query.pop();
			
			if (_t && _t.running) _t.stop();
			_t = vis.update(_dur);
			_t.play();
			
			exact = false; // reset exact match after each search
		}
	
		// --------------------------------------------------------------------
		
		/**
		 * Add interactivity controls: search box, time range adjuster,
		 * etc.
		 */
		private function addControls():void
		{
			// create search box
			search = new SearchBox(fmt, ">", 250);
			search.borderColor = 0xdedede;
			search.input.tabIndex = 0;
			search.addEventListener(SearchBox.SEARCH, onFilter);
			addChild(search);
			
			// create gender filter
			/*
			_gender = new Sprite(); _gender.name = "gender";
			_gender.addChild(new LegendItem("All", 0xff888888));
			_gender.addChild(new LegendItem("Male", 0xff8888ff));
			_gender.addChild(new LegendItem("Female", 0xffff8888))
			for (var i:int=0; i<_gender.numChildren; ++i) {
				var item:LegendItem = LegendItem(_gender.getChildAt(i));
				item.label.applyFormat(fmt);
				item.buttonMode = true;
				item.mouseChildren = false;
				item.margin = 3;
				item.render();
				if (i > 0) {
					var prev:LegendItem = LegendItem(_gender.getChildAt(i-1));
					item.x = prev.x + prev.width;
					item.alpha = 0.3;
				}
				item.addEventListener(MouseEvent.ROLL_OVER,
					function(e:MouseEvent):void {
						LegendItem(e.target).alpha = 1;
					}
				);
				item.addEventListener(MouseEvent.ROLL_OUT,
					function(e:MouseEvent):void {
						var item:LegendItem = LegendItem(e.target);
						if (item.text != _filter) item.alpha = 0.3;
					}
				);
				item.addEventListener(MouseEvent.CLICK,
					function(e:MouseEvent):void {
						for (var i:uint=0; i<_gender.numChildren; ++i) {
							LegendItem(_gender.getChildAt(i)).alpha = 0.3;
						}
						LegendItem(e.target).alpha = 1;
						_filter = LegendItem(e.target).text;
						onFilter();
					}
				);
			}
			addChild(_gender);
			*/
		}

		/**
		 * Place the visualization and controls.
		 */
		private function layout():void
		{	
			// Layout variables
			var yMargin:Number = 5;
			var visWidth:Number = 700;
			var visHeight:Number = 500;
			
			// Set the search box position
			search.x = 0;
			search.y = yMargin;
		
			// Set the visualization bounds
			_bounds = new Rectangle(0,0,visWidth,visHeight);
			_bounds.y += search.y + search.height + yMargin;
			vis.bounds = _bounds;
			trace(_bounds);
			vis.update();
		}

		// --------------------------------------------------------------------

		/**
		 * Reshapes a data set, pivoting from rows to _dates. For example, if
		 * yearly data is stored in individual rows, this method can be used to
		 * map each year into a column and the full time series into a single
		 * row. This is often needed to use the stacked area layout.
		 * @param tuples an array of data tuples
		 * @param cats the category values to maintain
		 * @param dim the dimension upon which to pivot. The values of this
		 *  property should correspond to the names of newly created _dates.
		 * @param measure the numerical value of interest. The values of this
		 *  property will be used as the values of the new _dates.
		 * @param cols an ordered array of the new column names. These should
		 *  match the values of the <code>dim</code> property.
		 * @param normalize a flag indicating if the data should be normalized
		 */
//			var data_rotated:Array =  reshape(rawData, ["artist"], "date", "playcount", _dates, false);
		public static function reshape(tuples:Array, cats:Array, dim:String,
			measure:String, cols:Array, normalize:Boolean=true):Array
		{
			var t:Object, d:Object, val:Object, name:String;
			var data:Array = [], names:Object = {};
			var totals:Object = {};
			for each (val in cols) totals[val] = 0;
			
			// create data set
			for each (t in tuples) {
				// create lookup hash for tuple
				var hash:String = "";
				for each (name in cats) hash += t[name];
				
				if (names[hash] == null) {
					// create a new data tuple
					data.push(d = {});
					for each (name in cats) d[name] = t[name];
					d[t[dim]] = t[measure];
					names[hash] = d;
				} else {
					// update an existing data tuple
					names[hash][t[dim]] = t[measure];
				}
				totals[t[dim]] += t[measure];
			}
			// zero out missing data
			for each (t in data) {
				var max:Number = 0;
				for each (name in cols) {
					if (!t[name]) t[name] = 0; // zero out null entries
					if (normalize)
						t[name] /= totals[name]; // normalize
					if (t[name] > max) max = t[name];
				}
				t.max = max;
			}
			
			return data;
		}
		
		private function printArray(a:Array):void {
			var i:Object, j:Object;
			var output:String = "";
			for(i in a)
			{
				for(j in a[i])
					output += a[i][j].toString() + " ";
				output += "\n";
			}
			trace(output);
		}

		private function fireEvent(type:String):void
		{
			Application.application.dispatchEvent(new Event(type));
		}

		/**
		 * Trims whitespace from start and end of string.
		 * Source: http://www.twisty.com/bandwagon/archives/2007/07/23/162642
		 */
		public static function trim(str:String):String
		{
    		for(var i:Number = 0; str.charCodeAt(i) < 33; i++);
    		for(var j:Number = str.length-1; str.charCodeAt(j) < 33; j--);
    		return str.substring(i, j+1);
		}
		
		override public function getExplicitOrMeasuredHeight():Number {
			return totalHeight;
		}

		override public function getExplicitOrMeasuredWidth():Number {
			return totalWidth;
		}
	}
}
